package meerkat.voting.controller.selector;

import meerkat.protobuf.Voting.*;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.List;
import java.lang.Math;

/**
 * A simple implementation of a QuestionSelector.
 * This implementation simply regards every single answer in the channel choice phase as an identifier of a category
 * Every category is an array of ballot questions.
 * Data of categories is initialized and stored by a SimpleCategoriesSelectionData protobuf.
 * After receiving the answers from a channel choice phase, this class simply gathers all the categories
 * chosen and compiles the list of ballot questions to include in the ballot for this voter (a question
 * is included if its index appears in any chosen category, or in the default category shared by all voters)
 */
public class SimpleListCategoriesSelector implements QuestionSelector {
    protected final static Logger logger = LoggerFactory.getLogger(SimpleListCategoriesSelector.class);

    private final BallotQuestion[] allBallotQuestions;
    private final int[] sharedDefaults;
    private final int[][][] categoryChoosers;

    private final static byte QUESTION_SELECTED = (byte)1;
    private final static byte QUESTION_NOT_SELECTED = (byte)0;



    public SimpleListCategoriesSelector(List<BallotQuestion> allBallotQuestions, SimpleCategoriesSelectionData data) {
        this.allBallotQuestions = new BallotQuestion[allBallotQuestions.size()];
        allBallotQuestions.toArray(this.allBallotQuestions);
        sharedDefaults = listToIntArray(data.getSharedDefaults().getQuestionIndexList());
        int[][][] selectionDataTmp = new int[data.getCategoryChooserList().size()][][];
        int channelChoiceQuestionNumber = 0;
        for (CategoryChooser catChooser: data.getCategoryChooserList()) {
            selectionDataTmp[channelChoiceQuestionNumber] = new int[catChooser.getCategoryList().size()][];
            int channelChoiceAnswerNumber = 0;
            for (Category category: catChooser.getCategoryList()) {
                selectionDataTmp[channelChoiceQuestionNumber][channelChoiceAnswerNumber] = listToIntArray(category.getQuestionIndexList());
                ++channelChoiceAnswerNumber;
            }
            ++channelChoiceQuestionNumber;
        }
        categoryChoosers = selectionDataTmp;
        assertDataValid();
    }

    private void assertDataValid () {
        int questionsLength = allBallotQuestions.length;
        int maxQuestionIndex = -1;
        for (int[][] categoryChooser: categoryChoosers) {
            for (int[] category: categoryChooser) {
                for (int index: category) {
                    maxQuestionIndex = Math.max(maxQuestionIndex, index);
                }
            }
        }
        if (maxQuestionIndex >= questionsLength) {
            String errorMessage = "Selection data refers to question index " + maxQuestionIndex + " while we have only " + questionsLength + " questions totally";
            logger.error(errorMessage);
            throw new IndexOutOfBoundsException(errorMessage);
        }
    }


    @Override
    public byte[] getChannelIdentifier(List<BallotAnswer> channelChoiceAnswers) {
        byte[] isSelected = new byte[allBallotQuestions.length];
        java.util.Arrays.fill(isSelected, QUESTION_NOT_SELECTED);

        for (int i: sharedDefaults) {
            isSelected[i] = QUESTION_SELECTED;
        }

        int channelChoiceQuestionNumber = 0;
        for (BallotAnswer ballotAnswer: channelChoiceAnswers) {
            assertAnswerLengthIsOne(ballotAnswer, channelChoiceQuestionNumber);
            for (int i: categoryChoosers[channelChoiceQuestionNumber][(int)ballotAnswer.getAnswer(0)]) {
                isSelected[i] = QUESTION_SELECTED;
            }
        }

        return isSelected;
    }

    private void assertAnswerLengthIsOne (BallotAnswer ballotAnswer, int questionNumber) {
        if (ballotAnswer.getAnswerCount() != 1) {
            String errorMessage = "SimpleListCategoriesSelector expects a single answer for every channel choice question\n";
            errorMessage += "Answer to question number " + (questionNumber+1) + " is";
            for (long i : ballotAnswer.getAnswerList()) {
                errorMessage += " " + i;
            }
            logger.error(errorMessage);
            throw new IllegalArgumentException(errorMessage);
        }
    }

    @Override
    public List<BallotQuestion> selectQuestionsForVoter(byte[] channelIdentifier) {
        List<BallotQuestion> selectedQuestions = new ArrayList<>();
        for (int i = 0; i < channelIdentifier.length; ++i) {
            if (channelIdentifier[i] == QUESTION_SELECTED) {
                selectedQuestions.add(allBallotQuestions[i]);
            }
        }
        return selectedQuestions;
    }

    private int[] listToIntArray(List<Integer> l) {
        int[] res = new int[l.size()];
        int index = 0;
        for (Integer i: l) {
            res[index] = i;
            ++index;
        }
        return res;
    }
}
